import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.sql.SQLException;
import java.sql.Connection;
import java.time.LocalDate;

import org.eclipse.rdf4j.IsolationLevel;
import org.eclipse.rdf4j.model.*;
import org.eclipse.rdf4j.repository.RepositoryConnection;

import org.eclipse.rdf4j.rio.RDFFormat;
import virtuoso.rdf4j.driver.*;


public class Test_Perf_RDF4J_1 extends Thread {

    public static final String VIRTUOSO_INSTANCE = "localhost";
    public static final String VIRTUOSO_PORT = "1111";
    public static final String VIRTUOSO_USERNAME = "dba";
    public static final String VIRTUOSO_PASSWORD = "dba";

    public static int max_threads = 10;
    public static boolean CommitAfterClear = false;

    public static String graph_name = "test:insert";
    public static String s_isolation = "repeatable_read";
    public static String s_concurrency = "default";

    public static String instance = VIRTUOSO_INSTANCE;
    public static String port = VIRTUOSO_PORT;
    public static String uid = VIRTUOSO_USERNAME;
    public static String pwd = VIRTUOSO_PASSWORD;

    public static int isolation = Connection.TRANSACTION_REPEATABLE_READ;
    public static int concurrency = VirtuosoRepository.CONCUR_DEFAULT;

    char pid;
    int id;
    int max_triples;

    ValueFactory vfac = null;
    IRI context = null;


    public Test_Perf_RDF4J_1(char _pid) {
        this.pid = _pid;
        this.max_triples = rnd_count();
    }

    //RND from 50 to 50000
    public static int rnd_count(){
        int max_rnd = 100;
        return (int) (Math.random() * max_rnd)*50;
    }


    public static int getIsolationLevel() {
       String s = s_isolation.toLowerCase();
       if (s.equals("read_uncommitted"))
         return Connection.TRANSACTION_READ_UNCOMMITTED;
       else if (s.equals("read_committed"))
         return Connection.TRANSACTION_READ_COMMITTED;
       else if (s.equals("repeatable_read"))
         return Connection.TRANSACTION_REPEATABLE_READ;
       else if (s.equals("serializable"))
         return Connection.TRANSACTION_SERIALIZABLE;
       else {
         s_isolation = "repeatable_read";
         return Connection.TRANSACTION_REPEATABLE_READ;
       }
    }

    public static int getConcurrencyMode() {
       String s = s_concurrency.toLowerCase();
       if (s.equals("optimistic"))
         return  VirtuosoRepository.CONCUR_OPTIMISTIC;
       else if (s.equals("pessimistic"))
         return VirtuosoRepository.CONCUR_PESSIMISTIC;
       else {
         s_concurrency = "default";
         return VirtuosoRepository.CONCUR_DEFAULT;
       }
    }


    public static void log(String mess) {
        System.err.println("   " + mess);
    }


    public static void main(String[] args) {

        if (args.length > 0)
          instance = args[0];

        if (args.length > 1)
          port = args[1];

        if (args.length > 2)
          uid = args[2];

        if (args.length > 3)
          pwd = args[3];

        if (args.length > 4)
          s_isolation = args[4];

        if (args.length > 5)
          s_concurrency = args[5];

        System.out.println("App options:");
        System.out.println(" % java Test_Perf_RDF4J hostname port_num uid pwd isolationMode concurrencyMode\n");
        System.out.println(" where ");
        System.out.println("   isolationMode: read_uncommitted | read_committed | repeatable_read | serializable ");
        System.out.println("                  default isolationMode = repeatable_read \n");
        System.out.println("   concurrencyMode: default | optimistic | pessimistic ");
        System.out.println("                  default concurrencyMode = default \n");

        System.out.println("Example of using:");
        System.out.println(" % java Test_Perf_RDF4J localhost 1111 dba dba repeatable_read default\n");

        System.out.println("===========================================================================\n");
        System.out.println("App will use next options");
        System.out.println("    hostname = "+instance);
        System.out.println("        port = "+port);
        System.out.println("         UID = "+uid);
        System.out.println("         PWD = "+pwd);
        System.out.println("   isolation = "+s_isolation);
        System.out.println(" concurrency = "+s_concurrency);
        System.out.println("===========================================================================\n");
        

        Test_Perf_RDF4J_1[] tests = new Test_Perf_RDF4J_1[max_threads];

        char pid = 'A';

        for(int i=0; i < max_threads; i++) {
          tests[i] = new Test_Perf_RDF4J_1((char)(pid+i));
        }

        for(Test_Perf_RDF4J_1 task : tests)
          task.start();

        try {
          for(Test_Perf_RDF4J_1 task : tests) {
            task.join();
          }
        } catch(Exception e) {
          log(e.toString());
        }
    }


    // RND from 1 to 100
    public static int rnd_skill(){
        int max_rnd = 100;
        return (int) (Math.random() * max_rnd);
    }

    public String genModel(int sz) {
        StringWriter out = new StringWriter(8192);
        try {
            int i = 0;

            String np1 = "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type> ";

            String np2 = "<http://www.test.com/lastLocationUpdate> ";
            String np3 = "<http://www.test.com/hasCountry> ";
            String np4 = "<http://www.test.com/hasSkill> ";

            String no1 = "<http://www.test.com/person> .\n";
            String no2 = "\""+ LocalDate.now() +"\"^^<http://www.w3.org/2001/XMLSchema#date> .\n";

            while(i < sz)
            {
                String ns = "<http://www.test.com/person_"+pid+"_"+id+"> ";
                id++;

                String no3 = "<http://www.test.com/country_"+pid+"> .\n";
                String no4 = "<http://www.test.com/skill_A"+rnd_skill()+"> .\n";
                String no5 = "<http://www.test.com/skill_B"+rnd_skill()+"> .\n";
                String no6 = "<http://www.test.com/skill_C"+rnd_skill()+"> .\n";
                String no7 = "<http://www.test.com/skill_D"+rnd_skill()+"> .\n";
                String no8 = "<http://www.test.com/skill_E"+rnd_skill()+"> .\n";
                String no9 = "<http://www.test.com/skill_F"+rnd_skill()+"> .\n";
                String no10 = "<http://www.test.com/skill_J"+rnd_skill()+"> .\n";

                out.write(ns); out.write(np1); out.write(no1); i++;
                out.write(ns); out.write(np2); out.write(no2); i++;
                out.write(ns); out.write(np3); out.write(no3); i++;
                out.write(ns); out.write(np4); out.write(no4); i++;
                out.write(ns); out.write(np4); out.write(no5); i++;
                out.write(ns); out.write(np4); out.write(no6); i++;
                out.write(ns); out.write(np4); out.write(no7); i++;
                out.write(ns); out.write(np4); out.write(no8); i++;
                out.write(ns); out.write(np4); out.write(no9); i++;
                out.write(ns); out.write(np4); out.write(no10); i++;
            }

        } catch (Exception e) {
            log("==["+Thread.currentThread().getName()+"]***FAILED Test " + e);
            return null;
        }

        return out.toString();
    }



    public void run() {

        RepositoryConnection conn = null;
        long i = 0;
        int sz = max_triples;

        try {
          VirtuosoRepository repository = new VirtuosoRepository("jdbc:virtuoso://" + instance + ":" + port, uid, pwd);

          repository.setConcurrencyMode(concurrency);
          vfac = repository.getValueFactory();
          context = vfac.createIRI(graph_name+"_"+pid);
          String base = "http://test";

          conn = repository.getConnection();
          ((VirtuosoRepositoryConnection)conn).setJdbcTransactionIsolation(isolation);

          log("==["+Thread.currentThread().getName()+"]== data prepared "+sz+"   graph name = "+context);
          String data = genModel(sz);

          while(true) {
                try {
                    log("==["+Thread.currentThread().getName()+"]== start TXN ");
                    conn.begin();
                    log("==["+Thread.currentThread().getName()+"]== clear graph "+context);
                    conn.clear(context);
                    if (CommitAfterClear) {
                      log("==["+Thread.currentThread().getName()+"]== commit after clear graph");
                      conn.commit();
                      log("==["+Thread.currentThread().getName()+"]== start TXN ");
                      conn.begin();
                    }
                    log("==["+Thread.currentThread().getName()+"]== start Add Data ");
                    conn.add((Reader)new StringReader(data), base, RDFFormat.NTRIPLES, context);
                    conn.commit();
                    log("==["+Thread.currentThread().getName()+"]== commit TXN ");
                } catch (Exception e) {
                    Throwable ex = e.getCause();
                    boolean deadlock = false;
                    if ((ex instanceof SQLException) && ((SQLException)ex).getSQLState().equals("40001"))
                        deadlock = true;

                    if (deadlock) {
                        log("==["+Thread.currentThread().getName()+"]== deadlock, rollback all and try insert again");
                        log("==["+Thread.currentThread().getName()+"]== rollback TXN ");
                        conn.rollback();
                        continue;
                    }
                    throw e;
                }

                break;
          }
          log("==["+Thread.currentThread().getName()+"]== data inserted "+sz);

        } catch (Exception e) {
            log("==["+Thread.currentThread().getName()+"]***FAILED Test " + e);
        } finally {
          if (conn != null)
            try {
              conn.close();
            } catch(Exception e) { }
        }

        log("==["+Thread.currentThread().getName()+"] DONE = "+i);
    }


}

